/*
   primf version 1.2  Copyright (C) 2003-2006 Frank Reglin

   primf is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   primf is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with primf; see the file COPYING.
*/

/* Diese Programm kann die Primzahlen bis 4000000000 berechnen und in einer 
   250 MByte grossen Datei abspeichern: 1 Bytes pro 16 Zahlen -- Es wird
   fuer jede ungerade Zahl ein Bit gespeichert, dass angibt, ob sie prim
   ist (1) oder nicht (0).

   index  0                           1
   bitnr  0  1  2  3  4  5  6 ... 31  0  1  2 . . .
   Zahl   1  3  5  7  9 11 13 ... 63 65 67 69 . . .

   Alle ungeraden Zahlen (z) lassen sich also auch als (index (i), bitnr (b))-Paar 
   darstellen: z = 64 * i + 2 * b + 1

   Fileformat "o8": nur ungerade(odd), jeweils 8 bits "umdrehen", um zur richtigen
   Bitreihenfolege zu kommen.
*/

#ifdef LINUX
 #define O_BINARY 0
 #include <errno.h>
#else
 #include <io.h>
#endif

#include <fcntl.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "mtime.h"



#define VERSION "1.2"

#define MAXNUMBER 4100000000U

/* die ersten 4 Bytes des Files zum Plausibilitaetstest = die Primzahlen
   bis 64. Rueckwaerts wegen Intel: die ersten 4 Bytes im File sind:
   6E CB B4 64   */
#define MAGIC 0x64B4CB6E 

static char * defaultfile = "primtab.o8";


/* 32 bit type: */
typedef unsigned int u32;

/* type for (index,bitno) representation: */
typedef struct
{
   int idx;
   int bit;
} pnum;


/* die Bits werden in einem statischen Array gehalten: */
u32 a[MAXNUMBER/64 + 2]; /* +2 wegen initialisierung in 3er-Blocks */

/* masks for bit operations: */
u32 mask[32] = 
{  0x00000001,  0x00000002,  0x00000004,  0x00000008, 
   0x00000010,  0x00000020,  0x00000040,  0x00000080, 
   0x00000100,  0x00000200,  0x00000400,  0x00000800, 
   0x00001000,  0x00002000,  0x00004000,  0x00008000,
   0x00010000,  0x00020000,  0x00040000,  0x00080000, 
   0x00100000,  0x00200000,  0x00400000,  0x00800000,
   0x01000000,  0x02000000,  0x04000000,  0x08000000, 
   0x10000000,  0x20000000,  0x40000000,  0x80000000 };
u32 negmask[32] = 
{  ~0x00000001, ~0x00000002, ~0x00000004, ~0x00000008, 
   ~0x00000010, ~0x00000020, ~0x00000040, ~0x00000080, 
   ~0x00000100, ~0x00000200, ~0x00000400, ~0x00000800, 
   ~0x00001000, ~0x00002000, ~0x00004000, ~0x00008000,
   ~0x00010000, ~0x00020000, ~0x00040000, ~0x00080000, 
   ~0x00100000, ~0x00200000, ~0x00400000, ~0x00800000,
   ~0x01000000, ~0x02000000, ~0x04000000, ~0x08000000, 
   ~0x10000000, ~0x20000000, ~0x40000000, ~0x80000000 };
   

int indexn( u32 num )
/* index for given number */
{
   return num >> 6; /* num/64 */
}


u32 number( int idx, int bit )
/* number from given (idx,bit):  64*idx + 2*bit + 1 */
{
   return ((unsigned)idx << 6)  +  (bit << 1)  + 1;
}


void cant( char * op, char * filename )
{
   printf( "FATAL: Cannot %s '%s' : %s\n", op, filename, strerror(errno) );
   exit( 2 );
}  


void print( u32 min, u32 max, char * ifile, int columns )
/* ifile == 0: print array for internal tests */
{
int maxi, fi, len, i, b, cnt;
u32 n, ai;
char * format;

   maxi = indexn( max );
   if( ifile )
   {
      fi = open( ifile, O_RDONLY | O_BINARY );
      if( fi < 0 )
         cant( "open (for reading)", ifile );
      
      len = read( fi, a, (maxi+1)*4 );
      if( len < 0 )
         cant( "read", ifile );

      if( (len >= 4) && (a[0] != MAGIC) )
      {
         printf( "FATAL: '%s' is not a prime table.\n", ifile );
         exit( 4 );
      }   
      if( len < (maxi+1)*4 )
      {
         len &= ~3; /* auf Vielfaches von 4 abrunden */
         if( len == 0 )
         {
             printf( "FATAL: '%s' is too short.\n", ifile );
             exit( 3 );
         }
         maxi = (len - 4) / 4;
         max = number( maxi, 31 ) + 1; 
         printf( "'%s' contains only the primes until %u\n", ifile, max );
         if( max < min )
            return;
      }
   }
   printf( "prime numbers from %u to %u\n", min, max );
   format = columns ? " %10u" : "%u\n";
   cnt = 0;
   if( min <= 2 )
   {
      printf( format, 2 );
      cnt++;
   }
   for( i = 0; i <= maxi; i++ )
   {
      ai = a[i];
      for( b = 0; b <= 31; b++ )
      {
         if( ai & mask[b] ) /* wenn nicht gestrichen ... */
         {
            n = number( i, b );
            if( (n >= min) && (n <= max) )
            {
               if( columns && (cnt >= columns) )
               {
                  printf( "\n" );
                  cnt = 0;
               }
               printf( format, number( i, b ) );
               cnt++;
            }
         }
      }
   }
   printf( "%send\n", columns ? "\n" : "" );
}


void calc( u32 max0, char * ofile )
{
u32 max;
int maxi, i, sqrti, fo, len, j;
pnum cand, p, del;
MTIME tim1, tim2;

   tim1 = mtime();

   /* auf 63 mod 64 aufrunden, um ganze Anzahl von 4Byte-Werten voll zu kriegen: */
   max = max0 | 63;
 
   maxi  = indexn( max );
   sqrti = indexn( (int)(sqrt(max)) + 1 );

   /* Initialisierung von a: Vielfache von 3 schon gestrichen */
   i = 0;
   for( j = 0; j <= maxi; j += 3 )
   {
      a[i++] = 0x6DB6DB6D;
      a[i++] = 0xDB6DB6DB;
      a[i++] = 0xB6DB6DB6;
   }

   a[0] ^= 3; /* ==> 1 keine Primzahl, aber 3 */

   /* initialisieren des Primzahlkandidaten: */
   cand.idx = 0; cand.bit = 2; /* start with 5 */
   
   /* Hier kommt nun der eigentliche Sieb-Algorithmus: */
   while( cand.idx <= sqrti ) /* fuer alle Zahlen bis zur Wurzel */
   {
      /* Ist die Zahl nicht gestrichen ? */
      if( (a[cand.idx] & mask[cand.bit]) )
      {   
         /* nicht gestrichen -> ist Primzahl: alle Vielfachen streichen: */
         p = cand;

         /* Die erste zu streichende Zahl ist p^2    !
            Die naechsten Zeilen sind deshalb: del := p^2 */
         del.bit = 2 * p.bit * (p.bit + 1);
         del.idx = 2 * p.idx * (32 * p.idx + 2 * p.bit + 1) + (del.bit >> 5);
         del.bit &= 31;
         // printf( "p = %u  p^2 = %u\n", number(p.idx,p.bit), number(del.idx,del.bit) );;;

         while( del.idx <= maxi )
         {
            /* Streichen: */
            a[del.idx] &= negmask[del.bit];
            /* naechste zu streichende Zahl: del := del + 2*p 
               2 * p damit sich wieder eine ungerade Zahl ergibt */
            del.bit += p.bit + p.bit + 1;
            del.idx += p.idx + p.idx + (del.bit >> 5);
            del.bit &= 31;
         }
      } /* if */
      /* naechste Zahl als Kandidat: */
      if( cand.bit == 31 )
      {
         cand.bit = 0;
         cand.idx++;
      }
      else
         cand.bit++;
   }

   tim2 = mtime();
   printf( "%u msec\n", (unsigned int)(tim2-tim1) );

   // print( 2, max0, 0, 7 );;;
   if( ofile )
   {
      fo = open( ofile, O_WRONLY | O_CREAT | O_BINARY, S_IREAD | S_IWRITE );
      if( fo < 0 )
         cant( "open (for writing)", ofile );
      len = write( fo, a, 4*(maxi+1) );
      if( len < 0 )
         cant( "write", ofile );
      close( fo );
   }
}


void usage( char * pgm )
{
   printf( "USAGE: %s [-p[cols] min] max [primefile]\n\n", pgm );
   printf( "       without -p: calculates all prime numbers <= max  and\n" );
   printf( "                   stores result in primefile [default=%s]\n\n", defaultfile );
   printf( "               -p: prints selected part of primefile\n" );
   printf( "                   2 <= min <= max <= %u\n", MAXNUMBER );
   printf( "                   cols = output columns [default=0 (flush left)]\n" );
   exit( 1 );
}


u32 getnumber( char * pgm, char * arg )
{
u32 num;

   if( (*arg < '0') || (*arg > '9') )
      usage( pgm );
   num = strtoul( arg, 0, 10 );
   if( (num < 2) || (num > MAXNUMBER) )
      usage( pgm );
   return num;
}


int main( int argc, char * argv[] )
{
u32 min, max;
char * primfile;
int cols;

   if( (argc < 2) || (argc > 5) )
      usage( argv[0] );
   printf( "primf  %s   (fr)\n", VERSION );
   if( (argv[1][0] == '-') && (argv[1][1] == 'p') )
   {   /* print */
      cols = atoi( &(argv[1][2]) );
      if( argc < 4 )
         usage( argv[0] );
      min = getnumber( argv[0], argv[2] );
      max = getnumber( argv[0], argv[3] );
      if( min > max )
         usage( argv[0] );
      primfile = argc==5 ? argv[4] : defaultfile;
      print( min, max, primfile, cols );
   }
   else
   {   /* calc */
      max = getnumber( argv[0], argv[1] );
      primfile = argc==3 ? argv[2] : defaultfile;
      calc( max, primfile );
   }
   return 0;
}


